#include <iostream.h>
#include <stdlib.h>
#include <math.h>

#include <time.h>
#include <sys/times.h>
#include <stdio.h>

/* Test Nullpunkt-in-Polygon fuer ein beliebiges doppelpunktfreies
   Polygon.

   Verwendung dieses Codes in eigenen Anwendungen nur mit
   ausdruecklicher Genehmigung des Autors.

   (C) Martin Ahlemeyer FIDO-NET 2:2432/1001.20
*/

typedef struct {
  int x,y;
} Punkt;

/*
  typedef struct {
  int x,y;
  } *Liste;
*/

typedef Punkt* Liste;

#define RAND 1  /* wenn der Rand zum Polygon gehoeren soll */

inline int Sector2(int x, int y) {
/* Ermittelt den Sektor, in dem (x,y) liegt */
/* 0 wenn (x,y) = (0,0) */
  return y>0 ? 1 : y<0 ? 2 : x>0 ? 1 : x<0 ? 2 : 0;
}

int InPoly(Liste L, int N) {
/* L[0] bis L[N] enthaelt die Punktliste des Polygons. L[0]=L[N] */

  int sold,s,count,i;
  long det;

  if ( !(s=Sector2(L[0].x,L[0].y)) )
    return RAND;  /* Sektor 0 */

  count = 0;

  for (i=1;i<=N;i++) {
    sold = s;
    if ( !(s=Sector2(L[i].x,L[i].y)) )
      return RAND;
    if ( s!=sold ) { /* Sektorwechsel */
      if ( !(det=(long)L[i-1].x*L[i].y-(long)L[i-1].y*L[i].x) )
        return RAND; /* Kante schneidet Nullpunkt */
      det>0 ? count++ : count--;
    }
  }
  return count!=0; /* count muss hier 2,0 oder -2 sein */
}

/*--- Die 4-Sektoren-Methode: ----------------------------*/

/* Test Nullpunkt-in-Polygon fuer ein beliebiges doppelpunktfreies
   Polygon.

   Verwendung dieses Codes in eigenen Anwendungen nur mit
   ausdruecklicher Genehmigung des Autors.

   (C) Martin Ahlemeyer FIDO-NET 2:2432/1001.20
*/

int Sector4(int x, int y) {
/* Ermittelt den Sektor, in dem (x,y) liegt */
/* 0 wenn (x,y) = (0,0) */

   if ( x>0 && y>=0 ) return 1;
   if ( y>0 && x<=0 ) return 2;
   if ( x<0 && y<=0 ) return 3;
   if ( y<0 && x>=0 ) return 4;
   return 0;

/* Effizientere Implementierung dieser Funktion vorbehalten ;-) */

}

int InPoly4(Liste L, int N) {
/* L[0] bis L[N] enthaelt die Punktliste des Polygons. L[0]=L[N] */

  int sold,s,count,i;
  long det;

  if ( !(s=Sector4(L[0].x,L[0].y)) )
    return RAND;  /* Sektor 0 */

  count = 0;

  for (i=1;i<=N;i++) {
    sold = s;
    if ( !(s=Sector4(L[i].x,L[i].y)) )
      return RAND; /* Sektor 0 */
    switch ( (4+s-sold)%4 ) {
      case 1: ++count;break; /* -> Linker Nachbarsektor */
      case 3: --count;break; /* -> Rechter Nachbarsektor */
      case 2: if ( !(det=(long)L[i-1].x*L[i].y-(long)L[i-1].y*L[i].x)
)
                return RAND; /* Kante schneidet Nullpunkt */
              det>0 ? count+=2 : (count-=2);
      /* Bei 0 tun wir nix, da kein Sektorwechsel */
    }
  }
  return count!=0; /* count muss hier 4,0 oder -4 sein */
}

int InPoly(Punkt p,Liste l,int n,int m) {
  Liste ll;
  int i;

  ll=new Punkt [n+1];
  memcpy(ll,l,sizeof(l));
  for (i=0;i<=n;i++) {
    ll[i].x-=p.x;
    ll[i].y-=p.y;
  }
  if (m==1)
    i=InPoly(ll,n);
  else
    i=InPoly4(ll,n);
  delete ll;
  return i;
}

int main() {
  int i,n;
  long c,cc;
  Liste l;
  Punkt p;
  double r,w;
  tms buf1,buf2;
  time_t t0,t1,t2,tm1,tm2;

  for (n=10;n<=400;n+=10) {
    cc=10000;
//  for(cc=100000;cc<=1000000;cc+=100000) {

    l=new Punkt [n+1];
    for (i=0;i<n;i++) {
      w=(double)i*2*M_PI/(double)n;
      r=(double)rand()/(double)RAND_MAX*100+10;
      l[i].x=(int)((double)r*cos(w));
      l[i].y=(int)((double)r*sin(w));
    }
    l[n]=l[0];
    
    w=(double)rand()/(double)RAND_MAX*2*M_PI;
    r=(double)rand()/(double)RAND_MAX*200;
    p.x=(int)((double)r*cos(w));
    p.y=(int)((double)r*sin(w));
    
    t1=times(&buf2);
    for (c=0;c<cc;c++);
    t2=times(&buf2);
    t0=t2-t1;
    
    t1=times(&buf2);
    for (c=0;c<cc;c++)
      InPoly(p,l,n,1);
    t2=times(&buf2);
    tm1=t2-t1-t0;
    
    t1=times(&buf2);
    for (c=0;c<cc;c++)
      InPoly(p,l,n,2);
    t2=times(&buf2);
    tm2=t2-t1-t0;
    
    /*
      cout << "Punkt liegt ";
      if (InPoly(p,l,n,1))
      cout << "innerhalb\n";
      else
      cout << "ausserhalb\n";
    */

    printf("%3d %10.8lf %10.8lf\n",n,
	   (double)tm1/100.0/(double)cc,
	   (double)tm2/100.0/(double)cc);

    delete l;
//  }
  }

  return 0;
}
