#include <math.h>
#include <vector.h>
#include <list.h>
#include <iostream.h>
#include <stdio.h>
extern "C" {
  #include <graphics.h>
}

//---------------------------------------------------------------------

class RGB:public vector {
  public:
    RGB();
    RGB(double lum,double r,double g,double b);
    double Luminance() { return get(0); }
    double Red() { return get(1); }
    double Green() { return get(2); }
    double Blue() { return get(3); }
};

RGB::RGB():vector() { }

RGB::RGB(double lum,double r,double g,double b)
    :vector(4,lum,r,g,b) {
  double n;
//  n=sqrt(r*r+g*g+b*b);
//  set(1,get(1)/n); set(2,get(2)/n); set(3,get(3)/n);
}

//---------------------------------------------------------------------

class Patch {
    vector x;
  public:
    Patch(vector x_);
    Patch(double x_,double y_,double z_);

    vector &X0() { return x; }
    virtual vector &X(double u,double v)=0;
    virtual vector &Normal(double u,double v)=0;
    virtual double A()=0;

    virtual RGB &Luminance(double u,double v)=0;
    virtual RGB &Reflex(double u,double v)=0;

    friend double Form(Patch &p1,Patch &p2);
};

Patch::Patch(vector x_) { x=x_; }

Patch::Patch(double x_,double y_,double z_) {
  x=vector(4,1,x_,y_,z_);
}

#define EPSILON 1e-10

double Form(Patch *p1,Patch *p2) {
  vector xi,xj,dij,dji,n1,n2;
  double f,r,rr,cw1,cw2;

  xi=p1->X(0.5,0.5); // the center of patch 1
  n1=p1->Normal(0.5,0.5);  
  xj=p2->X(0.5,0.5); // the center of patch 2
  n2=p2->Normal(0.5,0.5);  
  dij=xi-xj;
  dji=xj-xi;
  rr=scalar(dij,dij);
  r=sqrt(rr);
  if (r>=EPSILON) {
    cw1=scalar(dji,n1)/r;
    cw2=scalar(dij,n2)/r;
    f=cw1*cw2/rr;
  } else
    f=0;
  
  f=f/M_PI*p2->A();

  return f;
}

//---------------------------------------------------------------------

class PlanePatch:public Patch {
    vector e1,e2,n;
    double A0;
  public:
    PlanePatch(vector x_,vector p1,vector p2);
    PlanePatch(double x_,double y_,double z_,
	       double x1_,double y1_,double z1_,
	       double x2_,double y2_,double z2_);

    virtual vector &X(double u,double v);
    virtual vector &Normal(double u,double v);
    virtual double A();
};

PlanePatch::PlanePatch(vector x_,vector p1,vector p2)
           :Patch(x_) {

  e1=p1-X0();
  e2=p2-X0();
  n=e1*e2;
  A0=norm(n);
  n=n/A0;
}

PlanePatch::PlanePatch(double x_,double y_,double z_,
		       double x1_,double y1_,double z1_,
		       double x2_,double y2_,double z2_)
            :Patch(x_,y_,z_) {
  vector p1=vector(4,1,x1_,y1_,z1_);
  vector p2=vector(4,1,x2_,y2_,z2_);
  
  e1=p1-X0();
  e2=p2-X0();
  n=e1*e2;
  A0=norm(n);
  n=n/A0;
}

vector &PlanePatch::X(double u,double v) {
  return *(new vector(X0()+e1*u+e2*v));
}

vector &PlanePatch::Normal(double u,double v) { return n; }

double PlanePatch::A() { return A0; }

//-------------------------------------------------------

class WallPatch:public PlanePatch {
    RGB reflex;
    RGB emission;
  public:
    WallPatch(double x_,double y_,double z_,
	  double x1_,double y1_,double z1_,
	  double x2_,double y2_,double z2_,
	  double re,double ge,double be,double emit,
	  double rr,double gr,double br,double refl);
    WallPatch(vector x_,vector x1_,vector x2_,
	  double re,double ge,double be,double emit,
	  double rr,double gr,double br,double refl);

    virtual RGB &Luminance(double u,double v);
    virtual RGB &Reflex(double u,double v);
};

WallPatch::WallPatch(double x_,double y_,double z_,
		     double x1_,double y1_,double z1_,
		     double x2_,double y2_,double z2_,
		     double re,double ge,double be,double emit,
		     double rr,double gr,double br,double refl)
           :PlanePatch(x_,y_,z_,x1_,y1_,z1_,x2_,y2_,z2_) {

  emission=RGB(emit,re,ge,be); //--- emiting emit (absolut) in RGB
  reflex=RGB(refl,rr,gr,br); //--- reflecion % of incoming energy
}

WallPatch::WallPatch(vector x_,vector x1_,vector x2_,
		     double re,double ge,double be,double emit,
		     double rr,double gr,double br,double refl)
          :PlanePatch(x_,x1_,x2_) {

  emission=RGB(emit,re,ge,be); //--- emiting emit (absolut) in RGB
  reflex=RGB(refl,rr,gr,br); //--- reflecion % of incoming energy
}

//--- the luminance emitted by the patch

RGB &WallPatch::Luminance(double u,double v) {
  return emission;
}

RGB &WallPatch::Reflex(double u,double v) {
  return reflex;
}

//--------------------------------------------------------------------------------

class Plane {
  vector x,e1,e2;
  RGB reflex,emission;
  public:
    Plane(double x_,double y_,double z_,
	  double x1_,double y1_,double z1_,
	  double x2_,double y2_,double z2_,
	  double re,double ge,double be,double emit,
	  double rr,double gr,double br,double refl);
    virtual Patch **Patchlist(unsigned int nu,unsigned int nv);
};

Plane::Plane(double x_,double y_,double z_,
	     double x1_,double y1_,double z1_,
	     double x2_,double y2_,double z2_,
	     double re,double ge,double be,double emit,
	     double rr,double gr,double br,double refl) {

  x=vector(4,1,x_,y_,z_);
  vector p1=vector(4,1,x1_,y1_,z1_);
  vector p2=vector(4,1,x2_,y2_,z2_);
  
  e1=p1-x;
  e2=p2-x;

  emission=RGB(emit,re,ge,be); //--- emiting emit (absolut) in RGB
  reflex=RGB(refl,rr,gr,br); //--- reflecion % of incoming energy
}

Patch **Plane::Patchlist(unsigned int nu,unsigned int nv) {
  Patch **pl;
  unsigned int i,j,k;
  double u,v,du,dv;

  k=0;
  pl=new Patch* [nu*nv];
  du=1.0/(double)nu;
  dv=1.0/(double)nv;

  for (i=0;i<nv;i++) {
    v=(double)i*dv;
    for (j=0;j<nu;j++) {
      u=(double)j*du;
      pl[i*nu+j]=new WallPatch(x+e1*u+e2*v,x+e1*(u+du)+e2*v,x+e1*u+e2*(v+dv),
			      emission.Red(),emission.Green(),emission.Blue(),emission.Luminance(),
			      reflex.Red(),reflex.Green(),reflex.Blue(),reflex.Luminance());
    }
  }
  return pl;
}

//--------------------------------------------------------------------------------

int swap(double **A,double *b,unsigned int *index,unsigned int k,unsigned int n) {
  unsigned int i,j;
  double max;

  j=n+1;
  max=-1;

  for (i=k+1;i<n;i++) {
    if (fabs(A[i][k])>=max) {
      max=fabs(A[i][k]);
      j=i;
    }
  }
  if (max<EPSILON)
    return 0;
  i=index[k]; index[k]=index[j]; index[j]=i;
  for (i=0;i<n;i++) {
    max=A[k][i]; A[k][i]=A[j][i]; A[j][i]=max;
  }
  max=b[k]; b[k]=b[j]; b[j]=max;
  return 1;
}

int solve(double **A,double *x,double *b,unsigned int n) {
  unsigned int i,j,k;
  double pivot,s;
  unsigned int *index;

  index=new unsigned int [n];
  for (i=0;i<n;i++) index[i]=i;

  for (k=0;k<n;k++) {
    if (fabs(A[k][k])<EPSILON) {
      if (!swap(A,b,index,k,n))
	return 0;
    }
    for (i=k+1;i<n;i++) {
      pivot=A[i][k]/A[k][k];
      for (j=k+1;j<n;j++)
	A[i][j]-=pivot*A[k][j];
      b[i]-=pivot*b[k];
    }
  }

  k=1;
  for (i=n-1;k;i--) {
    s=b[i];
    for (j=i+i;j<n;j++)
      s-=A[i][j]*x[index[j]];
    x[index[i]]=s/A[i][i];
    k=i;
  }
  return 1;
}

//--------------------------------------------------------------------------------

int main() {
  //--- plane at (0,0,0) with normal (0,0,1), reflecting 40 %
  //    WallPatch wall(0,0,0, 1,0,0, 0,1,0, 0,0,0,0, 1,1,1,0.4);
  //--- plane at (0,0,1) emitting 100 (Lux ?), reflecting 10 %
  //    WallPatch roof(0,0,1, 1,0,0, 0,1,0, 1,1,1,100, 1,1,1,0.1);
  // cout << "Formfactor(wall,roof): " << Form(wall,roof) << "\n";
  // cout << "Formfactor(roof,wall): " << Form(roof,wall) << "\n";

  Plane wall(0,0, 0, 20,0,0, 0,20,0, 1,1,1,0  , 1,1,1,0.5);
  Plane roof(5,5,10, 0,10,0, 10,0,0, 1,1,1,100, 1,1,1,0.1);
  Patch **plwall,**plroof;
  List<Patch> pl;
  unsigned int i,j,np;
  
  double **form;
  double *B,*E,**A;
  double maxl;

  int drv,mode;
  vector pp;
  point p[5];

  np=2;

  plwall=wall.Patchlist(np,np);
  for(i=0;i<np*np;i++) 
    pl.Add(plwall[i]);

  plroof=roof.Patchlist(np,np);
  for(i=0;i<np*np;i++) 
    pl.Add(plroof[i]);

  cout << "Size of Patchlist: " << pl.Size() << "\n";
  cout << "Calculating form factors and matrix\n";

  form=new double* [pl.Size()];
  for (i=0;i<pl.Size();i++) {
    form[i]=new double [pl.Size()];
    for (j=0;j<pl.Size();j++) {
      if (i!=j)
	form[i][j]=-Form(pl.Get(i),pl.Get(j));
      else
	form[i][j]=0;
      printf("%+4lf ",form[i][j]);
    }
    printf("\n");
  }

  cout << "Done\n";
  cout << "Preparing Illumination model\n";

  A=new double* [pl.Size()];
  B=new double [pl.Size()];
  E=new double [pl.Size()];
  for (i=0;i<pl.Size();i++) {
    A[i]=new double [pl.Size()];
    for (j=0;j<pl.Size();j++) {
      A[i][j]=  pl.Get(i)->Reflex(0,0).Luminance()*pl.Get(i)->Reflex(0,0).Red()*form[i][j];
      if (i==j)
	A[i][j]=1+A[i][j];
      printf("%+4lf ",A[i][j]);
    }
    E[i]=pl.Get(i)->Luminance(0,0).Luminance()*pl.Get(i)->Luminance(0,0).Red();
    printf(" Bi = %4.3lf\n",E[i]);
  }

  cout << "Done\n";
  cout << "Calculation luminance vector (red)\n";

  solve(A,B,E,pl.Size());

  cout << "Done\n";

  maxl=-1;
  for (i=0;i<pl.Size();i++) {
    if (B[i]<0) B[i]=0;
    if (B[i]>maxl) maxl=B[i];
    printf("Patch[%4d].Red=%lf\n",i,B[i]);
  }

  drv=VGA;
  mode=VGAFREE;
  initgraph(&drv,&mode,"400*400*256");

  for (i=0;i<pl.Size();i++) {
    setrgbfillstyle(0,0,(int)(255.0*B[i]/maxl),
		    (int)(255.0*B[i]/maxl),(int)(255.0*B[i]/maxl));
    pp=pl.Get(i)->X(0,0);
    p[0].x=getmaxx()/2+(short)(pp.get(1)+pp.get(2)/sqrt(8))*5;
    p[0].y=getmaxy()/2-(short)(pp.get(3)+pp.get(2)/sqrt(8))*5;
    pp=pl.Get(i)->X(1,0);
    p[1].x=getmaxx()/2+(short)(pp.get(1)+pp.get(2)/sqrt(8))*5;
    p[1].y=getmaxy()/2-(short)(pp.get(3)+pp.get(2)/sqrt(8))*5;
    pp=pl.Get(i)->X(1,1);
    p[2].x=getmaxx()/2+(short)(pp.get(1)+pp.get(2)/sqrt(8))*5;
    p[2].y=getmaxy()/2-(short)(pp.get(3)+pp.get(2)/sqrt(8))*5;
    pp=pl.Get(i)->X(0,1);
    p[3].x=getmaxx()/2+(short)(pp.get(1)+pp.get(2)/sqrt(8))*5;
    p[3].y=getmaxy()/2-(short)(pp.get(3)+pp.get(2)/sqrt(8))*5;
    p[4].x=p[0].x;
    p[4].y=p[0].y;
    fillpoly(p,5);
    printf("%4d %4d %4d\n",p[0].x,p[0].y,(int)(255.0*B[i]/100.0));
  }

  flush();
  readkey();
  closegraph();

  return 0;
}

