#ifdef __cplusplus
extern "C" {
#endif
  #include <stdio.h>
  #include <stdlib.h>
#ifdef __cplusplus
}
#endif
  #include <math.h>

#include <vector.h>

static int icount=0;
const vector M_E1  =vector(4,0,1,0,0);
const vector M_E2  =vector(4,0,0,1,0);
const vector M_E3  =vector(4,0,0,0,1);
const vector M_E000=vector(4,0,0,0,0);
const vector M_E001=vector(4,0,0,0,1);
const vector M_E010=vector(4,0,0,1,0);
const vector M_E011=vector(4,0,0,1,1);
const vector M_E100=vector(4,0,1,0,0);
const vector M_E101=vector(4,0,1,0,1);
const vector M_E110=vector(4,0,1,1,0);
const vector M_E111=vector(4,0,1,1,1);

vector::vector(const vector &v) {
  int i;

  dim=v.dim;
  x=new double[dim];
  for (i=0;i<dim;i++) x[i]=v.x[i];
  #ifdef DEBUG
    icnt=icount++;
    printf("vector.o: vector::vector%d(%d,%lf,%lf,%lf,%lf)\n",icnt,dim,x[0],x[1],x[2],x[3]);
  #endif
}

vector::vector(int d,double x0,double x1,double x2,double x3) {
  x=NULL;

  #ifdef DEBUG
    icnt=icount++;
    printf("vector.o: vector::vector%d(%d,%lf,%lf,%lf,%lf)\n",icnt,d,x0,x1,x2,x3);
  #endif
  if (d>0) {
    x=new double[d];
    dim=d;
    x[0]=x0;
    if (d>=2) x[1]=x1;
    if (d>=3) x[2]=x2;
    if (d>=4) x[3]=x3;
  } else {
    #ifdef DEBUG
      printf("vector.o: warning in vector::vector 0 dimension\n");
    #endif
  }
}

double vector::get(const int i) const {
  #ifdef DEBUG
    printf("vector.o: vector::get%d(%d)\n",icnt,i);
  #endif
  if ((i>=0) && (i<dim))
    return x[i];
  else {
    printf("vector.o: error in vector::get index %d out of range for dimension %d\n",i,dim);
    return 0;
  }
}

void vector::set(int i,double v) {
  #ifdef DEBUG
    printf("vector.o: vector::set%d(%d,%lf)\n",icnt,i,v);
  #endif
  if ((i>=0) && (i<dim))
    x[i]=v;
  else {
    printf("vector.o: error in vector::set index %d out of range for dimension %d\n",i,dim);
  }
}

void vector::init(int d) {
  #ifdef DEBUG
    printf("vector.o: vector::init%d(%d)\n",icnt,d);
  #endif
  if (x==NULL) {
    dim=d;
    x=new double[d];
  } else
    printf("vector.o: warning in vector::init, already initialized\n");
}

vector &vector::operator=(const vector &v) {
  #ifdef DEBUG
    printf("vector.o: vector::operator%d=(%d)\n",icnt,v.icnt);
  #endif
  if (x!=NULL) delete x;
  dim=v.dim;
  x=new double[dim];
  for (int i=0;i<dim;i++) x[i]=v.x[i];
  return *this;  
}

vector operator+(const vector &x,const vector &y) {
  #ifdef DEBUG
    printf("vector.o: vector::operator%d+(%d)\n",x.icnt,y.icnt);
  #endif
  if (x.dim==y.dim) { 
    vector z;
    int i;

    z=x;
    for (i=1;i<x.dim;i++) z.x[i]=x.x[i]+y.x[i];
    z.x[0]=x.x[0];
    return vector(z);
  } else {
    printf("vector.o: error in vector::+ dimension mismatch\n");
  }
  return vector(4,0,0,0,0);
}

vector operator-(const vector &x,const vector &y) {
  #ifdef DEBUG
    printf("vector.o: vector::operator%d-(%d)\n",x.icnt,y.icnt);
  #endif
  if (x.dim==y.dim) { 
    vector z;
    int i;

    z=x;
    for (i=1;i<x.dim;i++) z.x[i]=x.x[i]-y.x[i];
    z.x[0]=x.x[0];
    return vector(z);
  } else {
    printf("vector.o: error in vector::- dimension mismatch\n");
  }
  return vector(4,0,0,0,0);
}

vector operator*(const vector &x,const vector &y) {

  #ifdef DEBUG
    printf("vector.o: vector::operator(%d)*(%d)vector\n",x.icnt,y.icnt);
  #endif
  if ((x.dim==4) && (y.dim==4)) { 
    vector z;

    z=x;
    z.x[1]=x.x[2]*y.x[3]-x.x[3]*y.x[2];
    z.x[2]=x.x[3]*y.x[1]-x.x[1]*y.x[3];
    z.x[3]=x.x[1]*y.x[2]-x.x[2]*y.x[1];
    z.x[0]=x.x[0]*y.x[0];
    return vector(z);
  } else {
    if (x.dim!=y.dim)
      printf("vector.o: error in vector::* dimension mismatch\n");
    else
      printf("vector.o: error in vector::* cross product only for dimension 3\n");
  }
  return vector(4,0,0,0,0);
}

vector operator*(double M[4][4],const vector &x) {
  vector z;
  int i,j;

  z=vector(x.dim);
  for (i=0;i<x.dim;i++) {
    z.x[i]=0;
    for (j=0;j<x.dim;j++) {
      z.x[i]=z.x[i]+M[i][j]*x.x[j];
    }
  }
//  z.x[0]=x.x[0];
  return vector(z);
}

vector vector::operator*(double y) const {
  #ifdef DEBUG
    printf("vector.o: vector::operator%d*%lf\n",icnt,y);
  #endif
  vector z=*this;
  int i;

  for (i=1;i<dim;i++) z.x[i]=x[i]*y;
  z.x[0]=x[0];
  return vector(z);
}

vector vector::operator/(double y) const {

  #ifdef DEBUG
    printf("vector.o: vector::operator%d/%lf\n",icnt,y);
  #endif
  vector z=*this;
  int i;

  for (i=0;i<dim;i++) z.x[i]=x[i]/y;
//  z.x[0]=x[0];
  return vector(z);
}

double norm(const vector &y) {
  double s;
  int i;

  s=0.0;
  #ifdef DEBUG
    printf("vector.o: norm((%d))\n",y.icnt);
  #endif
  for (i=1;i<y.dim;i++) s+=y.x[i]*y.x[i];
  s=sqrt(s);
  return s;
}

double scalar(const vector &x,const vector &y) {
  #ifdef DEBUG
    printf("vector.o: scalar((%d),(%d))\n",x.icnt,y.icnt);
  #endif
  if (x.dim==y.dim) { 
    double s=0.0;
    int i;

    for (i=1;i<x.dim;i++) {
      s+=x.get(i)*y.get(i);
    }
    return s;
  } else
    printf("vector.o: error in vector::scalar dimension mismatch\n");
  return 0.0;
}

int dimension(const vector &x) {
  return x.dim;
}

vector::~vector() {
  #ifdef DEBUG
    printf("vector.o: vector::~vector%d\n",icnt);
  #endif
  if (x!=NULL) delete x;
}

void printvec(const vector &v) {

  #ifdef DEBUG
    printf("%d: (%d)(",v.icnt,v.dim);
  #endif
  for (int i=1;i<v.dim;i++) {
    printf("%lf",v.x[i]);
    if (i<v.dim-1) printf(",");
  }
  printf(")\n");
}
/*
ostream &operator<<(ostream &str,const vector &v) {
  str << "(";
  str << v.dim << ",";
  for (int i=1;i<v.dim;i++) {
    str << v.x[i];
    if (i<v.dim-1) str << " ";
  }
  str << ")";
  return str; 
}
*/
vector rot(const vector &x,const vector &O,const vector &n,double phi) {
  vector x0,r,e,rr;
  double R;

  if (fabs(x.get(0))>1e-20) {
    x0=O+n*scalar(n,x-O);
    r=x-x0;
    r.set(0,0);
    R=norm(r);
    if (fabs(R)>1e-20) 
      r=r/R;
    else
      r=vector(4,1,0,0,0);
    e=(n*(-1))*r/norm(n);
    rr=x0+(r*cos(phi)+e*sin(phi))*R;
    rr.set(0,x.get(0));
  } else {
    r=x;
    e=n*r;
    rr=r*cos(phi)+e*sin(phi);
  }
  return vector(rr);
}

vector stdrot(const vector &x,const vector &O,const double u,const double v,const double w) {
  vector rr;

  rr=rot(x ,O,M_E1,u);
  rr=rot(rr,O,M_E2,v);
  rr=rot(rr,O,M_E3,w);
  return vector(rr);
}

